#include <unistd.h>
#include <stdio.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#ifdef USE_EMULATED_SYSTEM
# include "EmuDisplay.h"
#else
# include "RioDisplay.h"
#endif // USE_EMULATED_SYSTEM
#include "Window.h"

int Window::refcount = 0;
Window *Window::currentWindow = NULL;
Window *Window::lastWindow = NULL;
DisplayDev *Window::display = NULL;

Window::Window (Window *parent, const char *myName, int timeout) throw (int)
    : _name (myName), _parent (parent)
{
    if (parent != NULL)
        this->timeout = timeout;

    // Things to do on the first one..
    if (display == NULL) {
#ifdef USE_EMULATED_SYSTEM
        display = new EmuDisplay();
#else
        display = new RioDisplay();
#endif
        currentWindow = this;
        display->setPower (true);
    }

    mem = new char[display->memsize()];

    refcount++;
}

Window::~Window (void)
{
    if (--refcount) {
        shutdown();
    }
}

void
Window::shutdown (void)
{
    delete display;
    display = NULL;
}


void
Window::clear (void)
{
    memset (mem, DISPLAY_MEMSIZE, 0);
}

void
Window::refresh (void)
{
    if (currentWindow != NULL) {
        display->writeToScreen (currentWindow->mem, display->memsize());
        display->refresh ();
    }
}

/* Returns the x position that the write terminated at */
int
Window::writeString (int x, int y, font_t font, const char *text, bool reverse)
{
    int cur_x;
    int cur_y = y;
    int row;
    int max_x, max_y;
    int i, j;
    int on_mask, off_mask;

    if (x > DISPLAY_COLS || y > DISPLAY_ROWS) {
        return 0;
    }

    for (row = 1; row <= FONT_HEIGHT (font); row++) {
        cur_x = x;
        for (i = 0; i < strlen (text); i++) {
            uint32_t *font_char = font[(unsigned int)text[i]];
            for (j = 0; j < font_char[0]; j++) {
                char *this_loc = mem + (64*cur_y) + (cur_x/2);
                /* turn on pixel */
                if ((!reverse && font_char[row] & (1 << j)) ||
                    (reverse && !(font_char[row] & (1 << j)))) {
                    *this_loc |= (cur_x % 2) ? 0xf0 : 0x0f;
                } else {
                    *this_loc &= ~((cur_x % 2) ? 0xf0 : 0x0f);
                }
                cur_x++;

                /* Text overflow - break out of row loop. This will crop
                 * at the edge of the screen, but the next
                 * row will be written */
                if (cur_x > DISPLAY_COLS)
                    break;
            }
       }
       cur_y++;
       
       /* If we're off the bottom of the screen, bail out */
       if (cur_y > DISPLAY_ROWS)
        break;
    }
    return cur_x;
}

void
Window::fillRect (int x, int y, int width, int height, bool white)
{
    int cur_x, cur_y, max_x, max_y;

    max_x = (x + width) > DISPLAY_COLS ? DISPLAY_COLS : x + width;
    max_y = (y + height) > DISPLAY_ROWS ? DISPLAY_ROWS : y + height;

    for (cur_y = y; cur_y < max_y; cur_y++) {
        for (cur_x = x; cur_x < max_x; cur_x++) {
            char *this_loc = mem + (64*cur_y) + (cur_x/2);
            if (white) {
                /* turn off pixel */
                *this_loc &= ~((cur_x % 2) ? 0xf0 : 0x0f);
            } else {
                /* turn on pixel */
                *this_loc |= (cur_x % 2) ? 0xf0 : 0x0f;
            }
        }
    }
}


void
Window::drawRoundedRect (int x, int y, int width, int height, int thick, bool white)
{
    // Top Line
    fillRect (x + 1,         y,          width - 2, thick, white);
    // Bottom Line
    fillRect (x + 1,         y + height - thick, width - 2, thick, white);

    // Left Line
    fillRect (x,         y + 1,          thick, height - 2, white);
    // Right Line
    fillRect (x + width - thick,  y + 1,          thick, height - 2, white);
}

void
Window::drawRect (int x, int y, int width, int height, int thick, bool white)
{
    // Top Line
    fillRect (x,         y,          width, thick, white);
    // Bottom Line
    fillRect (x,         y + height - thick, width, thick, white);

    // Left Line
    fillRect (x,         y,          thick, height, white);
    // Right Line
    fillRect (x + width - thick, y,          thick, height, white);
}

void
Window::drawPixel (int x, int y, bool white)
{
    char *this_loc = mem + (64*y) + (x/2);
    if (white) {
        /* turn off pixel */
        *this_loc &= ~((x % 2) ? 0xf0 : 0x0f);
    } else {
        /* turn on pixel */
        *this_loc |= (x % 2) ? 0xf0 : 0x0f;
    }

}

void
Window::resetTimer (void)
{
    setTimer (timeout);
}

void
Window::setTimer (int sec)
{
    struct itimerval itv;
    itv.it_value.tv_sec = sec;
    itv.it_value.tv_usec = 0;
    signal (SIGALRM, alarmHandler);
    setitimer (ITIMER_REAL, &itv, NULL);
}

void
Window::clearTimer (void)
{
    setTimer (0);
}

void
Window::alarmHandler (int sig)
{
    struct itimerval itv;
    itv.it_value.tv_sec = 0;
    itv.it_value.tv_usec = 0;
    setitimer (ITIMER_REAL, &itv, NULL);

    switchToRoot ();
}

void
Window::switchTo (Window *window)
{
    display->lock();
    currentWindow->outFocus();
    lastWindow = currentWindow;
    currentWindow = window;
    currentWindow->inFocus();
    refresh();
    display->unlock();
}

void
Window::switchToRoot (void)
{
    display->lock ();
    currentWindow->outFocus ();
    lastWindow = currentWindow;
    while (currentWindow->_parent != NULL) {
        currentWindow = currentWindow->_parent;
    }
    currentWindow->inFocus ();
    refresh ();
    display->unlock ();
}

void
Window::switchToParent (void)
{
    display->lock();
    currentWindow->outFocus();
    lastWindow = currentWindow;
    currentWindow = _parent;
    currentWindow->inFocus();
    refresh();
    display->unlock();
}

void
Window::back (void)
{
    display->lock();
    Window *tmp = lastWindow;
    currentWindow->outFocus();
    lastWindow = currentWindow;
    currentWindow = tmp;
    currentWindow->inFocus();
    refresh();
    display->unlock();
}

int
Window::fontStrLen (font_t font, const char *str)
{
    int total = 0;
    for (int i = 0; i < strlen (str); i++) {
        uint32_t *font_char = font[(unsigned int)str[i]];
        total += font_char[0];
    }

    return total;
}

void
Window::drawTitledWindow (font_t font, const char *str)
{
    // Border
    drawRect (0, 0, DISPLAY_COLS, DISPLAY_ROWS, 1);

    // Title bar
    fillRect (0, 0, DISPLAY_COLS, FONT_HEIGHT (font) + 1);
    writeString ((DISPLAY_COLS - fontStrLen (font, str))/2, 1, font, str, true);
}
