#ifndef _WINDOW_H_
#define _WINDOW_H_
#include <stdio.h>
#include <iostream>
#include "DisplayDev.h"
#include "Thread.h"
#include "fonts.h"
#ifdef USE_EMULATED_SYSTEM
# include <sys/socket.h>
# include <linux/un.h>
#endif

/* The first occurence of a Window object will cause it to be set up */
/* The last removal of a Window object will tear it down again */

#define DISPLAY_MEMSIZE 4096
#define DISPLAY_ROWS 64
#define DISPLAY_COLS 128
#ifndef USE_EMULATED_SYSTEM
# define DISPLAY_DEVICE "/dev/display"
#else
# define DISPLAY_DEVICE "display"
#endif

#define DEFAULT_BUTTON_ACTION \
    fprintf (stderr, "%s called%s\n", __FUNCTION__, \
             stillPushed ? " (again)" : "");\
    return true;

class Window : public Thread
{
public:
    Window (Window *parent, const char *myName, int timeout=0) throw (int);
    ~Window (void);

public:
    int writeString (int x, int y, font_t font, const char *text,
                     bool reverse=false);
    void fillRect (int x, int y, int width, int height, bool white=false);
    void drawRect (int x, int y, int width, int height, int thick,
                   bool white=false);
    void drawRoundedRect (int x, int y, int width, int height, int thick,
                   bool white=false);
    void drawPixel (int x, int y, bool white=false);
    void drawTitledWindow (font_t font, const char *str);
    void switchToParent (void);
    static void back (void);
    static void switchTo (Window *window);
    static void switchToRoot (void);
    static int fontStrLen (font_t font, const char *str);

public:
    static bool onPowerButton (bool stillPushed)
                       { return currentWindow->_onPowerButton (stillPushed); };
    static bool onRepeatButton (bool stillPushed)
                       { return currentWindow->_onRepeatButton (stillPushed); };
    static bool onRandomButton (bool stillPushed)
                       { return currentWindow->_onRandomButton (stillPushed); };
    static bool onMenuButton (bool stillPushed)
                       { return currentWindow->_onMenuButton (stillPushed); };
    static bool onPlayButton (bool stillPushed)
                       { return currentWindow->_onPlayButton (stillPushed); };
    static bool onFFButton (bool stillPushed)
                       { return currentWindow->_onFFButton (stillPushed); };
    static bool onRewButton (bool stillPushed)
                       { return currentWindow->_onRewButton (stillPushed); };
    static bool onStopButton (bool stillPushed)
                       { return currentWindow->_onStopButton (stillPushed); };
    static bool onKnobLeft (bool stillPushed=false)
                       { return currentWindow->_onKnobLeft (stillPushed); };
    static bool onKnobRight (bool stillPushed=false)
                       { return currentWindow->_onKnobRight (stillPushed); };
    static bool onKnobButton (bool stillPushed)
                       { return currentWindow->_onKnobButton (stillPushed); };

private: /* implemented on a per-window basis */
    virtual void draw (void) {};
    virtual void inFocus (void) {};
    virtual void outFocus (void) {};
    virtual bool _onPowerButton (bool stillPushed)  { DEFAULT_BUTTON_ACTION };
    virtual bool _onRepeatButton (bool stillPushed) { DEFAULT_BUTTON_ACTION };
    virtual bool _onRandomButton (bool stillPushed) { DEFAULT_BUTTON_ACTION };
    virtual bool _onMenuButton (bool stillPushed)   { DEFAULT_BUTTON_ACTION };
    virtual bool _onPlayButton (bool stillPushed)   { DEFAULT_BUTTON_ACTION };
    virtual bool _onFFButton (bool stillPushed)     { DEFAULT_BUTTON_ACTION };
    virtual bool _onRewButton (bool stillPushed)    { DEFAULT_BUTTON_ACTION };
    virtual bool _onStopButton (bool stillPushed)   { DEFAULT_BUTTON_ACTION };
    virtual bool _onKnobLeft (bool stillPushed=false) { DEFAULT_BUTTON_ACTION };
    virtual bool _onKnobRight (bool stillPushed=false){ DEFAULT_BUTTON_ACTION };
    virtual bool _onKnobButton (bool stillPushed)   { DEFAULT_BUTTON_ACTION };

public:
    void clear (void);            // Clears display
    const char *name(void) const { return _name; };
    static void setPower (bool state) {if (display) display->setPower (state);}
    static void setTimer (int sec);
    static void clearTimer (void);
    static void alarmHandler (int sig);
    void resetTimer (void);

protected:
    char *mem;                           // Local copy of display RAM
    static void refresh (void);          // Updates the display output

private:
    int initialize (void);        // Sets up the display on first use
    void shutdown (void);         // Shuts down the display on last free

private:
    const char *_name;            // The name of this window
    static int refcount;          // How many windows are active
    Window *_parent;              // Parent window, NULL if it is root
    static Window *currentWindow; // Current window displayed
    static Window *lastWindow;    // Last window displayed
    int timeout;                  // Timeout before returning to root

protected:
    static DisplayDev *display;      // The hardware display
};
#endif // _WINDOW_H_
